/*
	Module: Serial DVD Controller for Philips/Pioneer Players	
	Copyright (C) 1999/2000  A. Schiffler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	To contact the author try:
	Andreas Schiffler                    aschiffler@home.com
	Senior Systems Engineer  -  Deskplayer.Com Inc., Buffalo
	4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada
	+1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)

*/

#include "dvdcontrol.h"

extern double timeGet();

/* ---------- DVD Command processor routines */

/* ------ Initialize DVD Player */

void DVDCommand_Init(struct dvdcontroller *dvd, int title)
{
 char aline[BUFFERLEN];

 if (dvd->enabled) {
  /* Store title */
  dvdcontroller_lock(dvd);
  dvd->title=title;
  dvdcontroller_unlock(dvd);
  /* Turn off on-screen display (OSD) */
  dvdcontroller_lock(dvd);
  if (dvd->type==DVD_PHILIPS) {
   strcpy(dvd->command,"[PC, OSD, OFF]\r");
  } else if (dvd->type==DVD_PIONEER) {
   strcpy(dvd->command,"0DS\r");
  } 
  dvdcontroller_unlock(dvd);
  /* Pause */
  usleep(20000);
  /* Set DVD playback title */
  dvdcontroller_lock(dvd);
  if (dvd->type==DVD_PHILIPS) {
   /* Nothing here */
  } else if (dvd->type==DVD_PIONEER) {
   sprintf(aline,"TI%iSE\r",title);
   strcpy(dvd->command,aline);
  } 
  dvdcontroller_unlock(dvd);
  /* Pause */
  usleep(20000);
 }
}

/* ------------ Send "Play Chapter X" Command to DVD */

void DVDCommand_Play(struct dvdcontroller *dvd, int chapter)
{
 char aline[BUFFERLEN];

 if (dvd->enabled) {
  /* Play specific title and chapter */
  dvdcontroller_lock(dvd);
  if (dvd->type==DVD_PHILIPS) {
   sprintf(aline,"[PC,PlayC,%i,%i]\r",dvd->title,chapter);
   strncpy(dvd->command,aline,BUFFERLEN);
  } else if (dvd->type==DVD_PIONEER) {
   sprintf(aline,"CH%iSE PL\r",chapter);
   strncpy(dvd->command,aline,BUFFERLEN);
  } 
  dvdcontroller_unlock(dvd);
 }
}

/* ------- Read DVD Result code (no processing or checking done) */

void ParseDVDResult(struct dvdcontroller *dvd, char *line, double commandtime)
{
 float delay;

 if (commandtime>0.0) {
  delay=(float)(timeGet()-commandtime);
 } else {
  delay=0.0;
 }

#ifdef DEBUG
 fprintf (stderr,"%s-DVD on %s said (%5.2f sec):%s\n",dvd_names[dvd->type],dvd->device,(float)delay,line);
 fflush (stderr);
#endif

}

/* -------- DVD Controller Thread */

int ControllerThread(struct dvdcontroller *dvd)
{
 int i;
 int inloop;
 int fd;
 int result;
 char buffer[BUFFERLEN];
 char *curbuffer;
 char linebuffer[BUFFERLEN];
 char *curlinebuffer;
 int linebufferpos;
 struct timeval timeout;
 int maxfd;
 fd_set readfs;
 double commandtime=0.0;

 /* Get local filehandle */
 dvdcontroller_lock(dvd);
 fd=dvd->fd;
 dvdcontroller_unlock(dvd);
 
 /* Serial-port processing loop */
 inloop=1;
 maxfd=fd+1;
 linebufferpos=0;
 curlinebuffer=linebuffer;
 while (inloop) {
  FD_SET(fd, &readfs);
  /* set timeout value within input loop */
  timeout.tv_usec = 10;  /* milliseconds */       
  timeout.tv_sec  = 0;   /* seconds */       
  /* Wait for data */
  result = select(maxfd, &readfs, NULL, NULL, &timeout);
  if (result>0) {
   /* Have data, read it */
   result = read(fd,buffer,BUFFERLEN);
   buffer[result]=0;
   /* Transfer data into linebuffer */
   curbuffer=buffer;
   for (i=0; i<result; i++) {
    /* Process buffer */
    if ((*curbuffer=='\n') || (*curbuffer=='\r')) {
     /* Line is done */
     *curlinebuffer=0;
     /* Process content */
     ParseDVDResult(dvd,linebuffer,commandtime);
     /* Reset line */
     curlinebuffer=linebuffer;
     linebufferpos=0;     
    } else {
     /* Add character to line */
     *curlinebuffer=*curbuffer;
     curlinebuffer++;
     linebufferpos++;
    }
    /* Check line length */
    if (linebufferpos==BUFFERLEN) {
     /* Line is full */
     *curlinebuffer=0;
     /* Process content */
     ParseDVDResult(dvd,linebuffer,commandtime);
     /* Reset line */
     curlinebuffer=linebuffer;
     linebufferpos=0;     
    };
    curbuffer++;
   }
  }
  /* Get new command and copy to local buffer, clear command buffer */
  dvdcontroller_lock(dvd);
  if (strlen(dvd->command)>0) {
#ifdef DEBUG
   /* Debug output of command */
   fprintf (stderr,"\nSend %s-DVD on %s:%s\n",dvd_names[dvd->type],dvd->device,dvd->command);
   fflush(stderr);
#endif
   /* Copy command to transfer buffer */
   strncpy(buffer,dvd->command,BUFFERLEN);
   dvd->command[0]=0;
   dvdcontroller_unlock(dvd);
   /* Send DVD command */
   commandtime=timeGet();
   result=write(fd,buffer,strlen(buffer));
  } else {
   dvdcontroller_unlock(dvd);
  }
 }
 
 return(0);
}

/* -------- Thread locking routines */

int dvdcontroller_lock(struct dvdcontroller *dvd)
{
 /* Lock shared variables */ 
 pthread_mutex_lock(&dvd->dvdcontrol_mutex);
 
 return(0);
}

int dvdcontroller_unlock(struct dvdcontroller *dvd)
{
 /* Unlock shared variables */ 
 pthread_mutex_unlock(&dvd->dvdcontrol_mutex);
 
 return(0);
}

/* --------- Start DVD controller thread */

int start_dvdcontroller (struct dvdcontroller *dvd)
{
 pthread_t reader_thread;

 if (dvd->enabled) {
  /* Start thread */
  if (pthread_create(&reader_thread, NULL,(void *)ControllerThread,(void *)dvd)==-1) {
   perror ("pthread_create()");
   return(-1);
  } else {
   /* Store thread handle */
   dvd->thread=reader_thread;
   return(0);
  }
 } else {
  fprintf (stderr,"start_dvdcontroller: cannot start - device not enabled.\n");
  return(-1);
 }
}

/* -------- Initialize device and thread */

int init_dvdcontroller(struct dvdcontroller *dvd, char* device, int type)
{
 struct termios newtio;

 /* Turn it off */
 dvd->enabled=0;

 /* Initialize variables */
 dvd->thread=-1;
 strncpy(dvd->device,device,BUFFERLEN);
 strcpy(dvd->command,"");
 dvd->type=type;

 /* Initialize sycro mutex */
 if (pthread_mutex_init(&dvd->dvdcontrol_mutex, NULL)==-1) {
  perror("pthread_mutex_init()");
  return(-1);
 }

 /* Check type */
 if (type==DVD_NONE) {
  return(0);
 }
 
 /* Open device and set type */
 dvd->fd = open(device, O_RDWR | O_NOCTTY );
 if (dvd->fd<0) {
  perror("open()"); 
  return(-1); 
 }

 /* Termial setup */
 tcgetattr(dvd->fd,&dvd->oldtio); 
 bzero(&newtio, sizeof(newtio));
 if (type==DVD_PHILIPS) { 
  newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
 } else if (type==DVD_PIONEER) {
  newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
 }
 newtio.c_iflag = IGNPAR;
 newtio.c_oflag = 0;
 newtio.c_lflag = 0;
 newtio.c_cc[VTIME]    = 0;
 newtio.c_cc[VMIN]     = 1;
 tcflush(dvd->fd, TCIFLUSH);
 tcsetattr(dvd->fd,TCSANOW,&newtio);
 
 /* Enable device */
 dvd->enabled=1; 
 return(0);
}

/* --------- Stop DVD thread */

int stop_dvdcontroller (struct dvdcontroller *dvd)
{

 if (dvd->enabled) {
  /* Shutdown thread */
  if (dvd->thread!=-1) {
   pthread_cancel(dvd->thread);
  }

  /* Restore the old port settings */
  tcsetattr(dvd->fd,TCSANOW,&dvd->oldtio);

  /* Close device */
  close(dvd->fd);
 }

 return(0);
}
