
/*
	Module: Read Joystick Positions from A/D Voltage Values
	Copyright (C) 1999/2000  A. Schiffler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	To contact the author try:
	Andreas Schiffler                    aschiffler@home.com
	Senior Systems Engineer  -  Deskplayer.Com Inc., Buffalo
	4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada
	+1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)

*/

#include "joyread.h"

/* -------- Get a adc voltage reading */

float readADCVoltage (struct joyreader *joy, int channel)
{
 float value;

 adcreader_lock(joy->adc);
 value=joy->adc->values[channel]; 
 adcreader_unlock(joy->adc);
 
 return(value);
}

/* -------- Serial A/D Converter Reader Thread with Conversion */

int JoyReaderThread(struct joyreader *joy)
{
 int i;
 float newx=0.0, newy=0.0;
 float newu,newv,diffu,diffv;
 int inloop;
 int localtestmode;
 int localchannelx;
 int localchannely;
 double localxfactor;
 double localyfactor;
 float minu,maxu,minv,maxv;
 float bufferu[NUM_AVERAGE],bufferv[NUM_AVERAGE];
 int bufferpos=0;
 float sumu=0.0;
 float sumv=0.0;

 /* Initialize calibration boundaries */
 minu=1.0;
 maxu=0.0;
 minv=1.0;
 maxv=0.0;
 /* Initialize buffers */
 for (i=0; i<NUM_AVERAGE; i++) {
  bufferu[i]=0.0;
  bufferv[i]=0.0;
 }
 /* Go into processing loop */
 inloop=1;
 while (inloop) {
  /* Determine joystick mode, channels and calibration data  */
  joyreader_lock(joy);
  localtestmode=joy->testmode;
  localchannelx=joy->channelx;
  localchannely=joy->channely;
  localxfactor=joy->xfactor;
  localyfactor=joy->yfactor;
  minu=joy->minu;
  maxu=joy->maxu;
  minv=joy->minv;
  maxv=joy->maxv;
  joyreader_unlock(joy);
  /* Process according to mode */
  if (localtestmode==JOY_MANUAL) { 
   /* ---------- No processing ---------------- */
   usleep(1000);
  } else if ((localtestmode==JOY_CALIBRATE) || (localtestmode==JOY_NORMAL)) {
   /* ----------- Calibration processing ---------- */
   /* Read voltages */
   newu=readADCVoltage(joy,localchannelx);
   newv=readADCVoltage(joy,localchannely);
   /* Store values in buffer and calculate moving-average */
   /* Update buffer position */
   bufferpos++; 
   if (bufferpos==NUM_AVERAGE) bufferpos=0;
   /* Update sum for u */
   sumu -= bufferu[bufferpos];
   sumu += newu;
   bufferu[bufferpos]=newu;
   /* Update sum for v */
   sumv -= bufferv[bufferpos];
   sumv += newv;
   bufferv[bufferpos]=newv;
   /* Calculate moving average */
   newu=sumu*(1.0/NUM_AVERAGE);
   newv=sumv*(1.0/NUM_AVERAGE);
   if (localtestmode==JOY_CALIBRATE) {
    /* Adjust minima and maxima */
    if (minu>newu) {
     minu=newu;
    } else if (maxu<newu) {
     maxu=newu;
    }
    if (minv>newv) {
     minv=newv;
    } else if (maxv<newv) {
     maxv=newv;
    }
    /* Store maxima */
    joyreader_lock(joy);
    joy->minu=minu;
    joy->maxu=maxu;
    joy->minv=minv;
    joy->maxv=maxv;
    joyreader_unlock(joy);
   }
   /* Calulate corrected coordinates */
   diffu=maxu-minu; 
   if (diffu==0.0) {
    newx=newu;
   } else {
    newx=(newu-minu)/diffu;
   }
   diffv=maxv-minv; 
   if (diffv==0.0) {
    newy=newv;
   } else {
    newy=(newv-minv)/diffv;
   }
   /* Calibrate for nonlinearities */
   if (newx<0.0) {
    newx=0.0;
   } else if (newx>1.0) {
    newx=1.0;
   } 
   if (newy<0.0) {
    newy=0.0;
   } else if (newy>1.0) {
    newy=1.0;
   } 
   newx = (float)pow((double)newx,localxfactor);
   newy = (float)pow((double)newy,localyfactor);
   /* Scale to -1..1 range, inverting X */
   newx =  1.0-2.0*newx;
   newy = -1.0+2.0*newy;
   if (newx<-1.0) {
    newx=-1.0;
   } else if (newx>1.0) {
    newx=1.0;
   } 
   if (newy<-1.0) {
    newy=-1.0;
   } else if (newy>1.0) {
    newy=1.0;
   } 
   /* Store coordinates */
   joyreader_lock(joy);
   joy->x=newx;
   joy->y=newy; 
   joyreader_unlock(joy);
  }
  /* Pause */
  usleep(10000);
 }
 
 return(0);
}

/* --------- Thread locking routines */

int joyreader_lock(struct joyreader *joy)
{
 /* Lock shared variables */ 
 pthread_mutex_lock(&joy->joyread_mutex);
 
 return(0);
}

int joyreader_unlock(struct joyreader *joy)
{
 /* Unlock shared variables */ 
 pthread_mutex_unlock(&joy->joyread_mutex);
 
 return(0);
}

int start_joyreader (struct joyreader *joy)
{
 pthread_t reader_thread;

 /* Start thread */
 if (pthread_create(&reader_thread, NULL,(void *)JoyReaderThread,(void *)joy)==-1) {
  perror ("pthread_create()");
  return(-1);
 } else {
  joy->thread=reader_thread;
  return(0);
 }
}

/* -------- Initialize data structures and thread */

int init_joyreader(struct joyreader *joy, struct adcreader *adc, int channelx, int channely)
{
 /* Initialize variables */
 joy->thread=-1;
 joy->testmode=JOY_NORMAL;
 joy->adc=adc;
 joy->channelx=channelx;
 joy->channely=channely;
 joy->x=0.0;
 joy->y=0.0;
 joy->minu=0.0;
 joy->maxu=1.0;
 joy->minv=0.0;
 joy->maxv=1.0;
 joy->xfactor=1.0;
 joy->yfactor=0.65;
 
 /* Initialize sycro mutex */
 if (pthread_mutex_init(&joy->joyread_mutex, NULL)==-1) {
  perror("pthread_mutex_init()");
  return(-1);
 }

 return(0);
}

/* -------- Stop thread */

int stop_joyreader (struct joyreader *joy)
{
 /* Shutdown thread */
 if (joy->thread!=-1) {
  pthread_cancel(joy->thread);
 }

 return(0);
}
