
/*
	Module: Serial Relais Board Interface to Control Lights
	Copyright (C) 1999/2000  A. Schiffler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	To contact the author try:
	Andreas Schiffler                    aschiffler@home.com
	Senior Systems Engineer  -  Deskplayer.Com Inc., Buffalo
	4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada
	+1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)

*/

#include "lights.h"

/* ------ Routine to get the time as float */

extern double timeGet();

/* ------- Control lights through serial port */

int SerialThread(struct lightcontroller *light)
{
 int i;
 int inloop;
 int fd;
 int result;
 struct timeval timeout;
 int maxfd;
 fd_set readfs;
 unsigned char buffer[4];
 unsigned char value, oldvalue;
 double testtime;
   
 /* Get local filehandle */
 lightcontroller_lock(light);
 fd=light->fd;
 lightcontroller_unlock(light);

 /* --------- Initialize board */
 /* set timeout value within input loop */
 timeout.tv_usec = 10;  /* milliseconds */       
 timeout.tv_sec  = 0;   /* seconds */       
 /* ---- write init command ---- */
 buffer[0]=1;
 buffer[1]=1;
 buffer[2]=0;
 buffer[3]=buffer[0];
 buffer[3] ^= buffer[1];
 buffer[3] ^= buffer[2];
 result=write(fd,buffer,4);
 /* Wait for data */
 result = select(maxfd, &readfs, NULL, NULL, &timeout);
 if (result>0) {
  /* Have data, read it */
  result = read(fd,buffer,4);
 }
  
 /* Pause */
 usleep(100000);

 /* Setup a period for testing */
 testtime = timeGet()+5.0;
 
 /* Serial-port processing loop */
 inloop=1;
 maxfd=fd+1;
 value=0;
 oldvalue=1;
 while (inloop) {
  /* Check testime */
  if (timeGet()<testtime) {
   /* Turn all lights on */
   value=255-1;
  } else {
   /* Get new value from thread */
   lightcontroller_lock(light);
   value=light->value;
   lightcontroller_unlock(light);
  }
  
  /* set timeout value within input loop */
  FD_SET(fd, &readfs);
  timeout.tv_usec = 10;  /* milliseconds */       
  timeout.tv_sec  = 0;   /* seconds */       
  
  /* Check for change */
  if (oldvalue==value) {
   /* ---- write a setting ---- */
   buffer[0]=0;
   buffer[1]=1;
   buffer[2]=0;
   buffer[3]=buffer[0];
   buffer[3] ^= buffer[1];
   buffer[3] ^= buffer[2];
   result=write(fd,buffer,4);
  } else {
   /* Value has changed - send new setting */
   /* ---- write a setting ---- */
   buffer[0]=3;
   buffer[1]=1;
   buffer[2]=value;
   buffer[3]=buffer[0];
   buffer[3] ^= buffer[1];
   buffer[3] ^= buffer[2];
   result=write(fd,buffer,4);
   /* Remember */
   oldvalue=value;
  }
  /* Wait for data */
  result = select(maxfd, &readfs, NULL, NULL, &timeout);
  if (result>0) {
   /* Have data, read it */
   result = read(fd,buffer,4);
  }
  
  /* Throttle rate */
  usleep(1000000);
  
  value++;
 }

 /* Pause */
 usleep(100000);
 
 return(0);
}

/* -------- Initialize by clearing */

int ClearSwitches(struct lightcontroller *light)
{
 int i;
 int inloop;
 int fd;
 int result;
 struct timeval timeout;
 int maxfd;
 fd_set readfs;
 unsigned char buffer[4];
 unsigned char value;
 
 /* Get local filehandle */
 lightcontroller_lock(light);
 fd=light->fd;
 lightcontroller_unlock(light);

 /* Initialize board */
 /* set timeout value within input loop */
 timeout.tv_usec = 10;  /* milliseconds */       
 timeout.tv_sec  = 0;   /* seconds */       

 /* Loop 3 times */
 for (i=0; i<3; i++) {
 
  FD_SET(fd, &readfs);
  /* set timeout value within input loop */
  timeout.tv_usec = 10;  /* milliseconds */       
  timeout.tv_sec  = 0;   /* seconds */       

  /* ---- WRITE section ---- */
  buffer[0]=3;
  buffer[1]=1;
  buffer[2]=0;
  buffer[3]=buffer[0];
  buffer[3] ^= buffer[1];
  buffer[3] ^= buffer[2];
  result=write(fd,buffer,4);

  /* ---- READ section ----- */
  /* Wait for data */
  result = select(maxfd, &readfs, NULL, NULL, &timeout);
  if (result>0) {
   /* Have data, read it */
   result = read(fd,buffer,4);
  }
 
  /* Throttle rate */
  usleep(100000);
 }
 
 return(0);
}

/* --------- Thread locking routines */

int lightcontroller_lock(struct lightcontroller *light)
{
 /* Lock shared variables */ 
 pthread_mutex_lock(&light->lightcontrol_mutex);
 
 return(0);
}

int lightcontroller_unlock(struct lightcontroller *light)
{
 /* Unlock shared variables */ 
 pthread_mutex_unlock(&light->lightcontrol_mutex);
 
 return(0);
}

/* -------- Start controller thread */

int start_lightcontroller (struct lightcontroller *light)
{
 pthread_t reader_thread;

 /* Check if enabled */
 if (light->enabled) {
  /* Store parameters */
  lightcontroller_lock(light);
  lightcontroller_unlock(light);
  /* Start thread */
  if (pthread_create(&reader_thread, NULL,(void *)SerialThread,(void *)light)==-1) {
   perror ("pthread_create()");
   return(-1);
  } else {
   light->lightcontrol_thread=reader_thread;
   return(0);
  }
 } else {
  fprintf (stderr,"start_lightcontroller: cannot start thread - device not enabled.\n");
  return(-1);
 }
}

/* ------- Initialize device and thread */

int init_lightcontroller(struct lightcontroller *light, char *device, int type)
{
 int i,j;
 struct termios newtio;

 /* Turn off */
 light->enabled=0;

 /* Initialize sycro mutex */
 if (pthread_mutex_init(&light->lightcontrol_mutex, NULL)==-1) {
  perror("pthread_mutex_init()");
  return(-1);
 }

 /* Initialize variables */
 strncpy(light->device,device,LIGHTBUFFERLEN);

 /* Check type */
 if (type==LIGHT_NONE) {
  return(0);
 }
  
 /* Open device and set type */
 light->fd = open(device, O_RDWR | O_NOCTTY );
 if (light->fd<0) {
  perror("open()"); 
  return(-1); 
 }

 /* Termial setup */
 tcgetattr(light->fd,&light->oldtio); 
 bzero(&newtio, sizeof(newtio));
 newtio.c_cflag = B19200 | CS8 | CLOCAL | CREAD;
 newtio.c_iflag = IGNPAR;
 newtio.c_oflag = 0;
 newtio.c_lflag = 0;
 newtio.c_cc[VTIME]    = 0;
 newtio.c_cc[VMIN]     = 1;
 tcflush(light->fd, TCIFLUSH);
 tcsetattr(light->fd,TCSANOW,&newtio);
 
 /* Turn on */
 light->enabled=1;
 return(0);
}

/* -------- Shutdown thread and reset relais */

int stop_lightcontroller (struct lightcontroller *light)
{
 if (light->enabled) {
  /* Blank switches */
  
  /* Shutdown thread */
  if (light->lightcontrol_thread!=-1) {
   pthread_cancel(light->lightcontrol_thread);
  }

  /* Turn all lights off */
  ClearSwitches(light);
  
  /* Restore the old port settings */
  tcsetattr(light->fd,TCSANOW,&light->oldtio);

  /* Close device */
  close(light->fd);
 }

 return(0); 
}
