/*

    Subroutine: Utility routines using XWindows access through shared memory.
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer    -    Deskplayer Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

#include "xutil.h"

#include "font8x8.h"

/* Additional colormap support my M. Hoch */

int setup_xwinbuffer(struct xwinbuffer *xwin, char *prgname, int width, int height, int depth) {
  int screen;
  XWindowAttributes win_attributes;
  XVisualInfo desired_visual_template;
  XVisualInfo *visual_info_list; 
  int no_visual_matched;
  XGCValues gc_values;
  int localdepth;
  Colormap   wtcolormap ;
  XColor   * palette;
  int        pixelNum, numCols;

  /* Store parameters */
  xwin->width=width;
  xwin->height=height;
  xwin->depth=depth; 

  /* Reset mode-switch variables */ 
  xwin->modeswitch=0;
  xwin->vm_smode=-1;

  /* Calculate bounds */
  xwin->maxx=width-1;
  xwin->maxy=height-1;
  
  /* Open window */
  xwin->display_ = XOpenDisplay(NULL);
  xwin->wtwin_ = XCreateSimpleWindow(xwin->display_, DefaultRootWindow(xwin->display_),
			       0, 0, xwin->width, xwin->height, xwin->depth,
			       BlackPixel(xwin->display_, DefaultScreen(xwin->display_)),
			       BlackPixel(xwin->display_, DefaultScreen(xwin->display_)));  
  XMapWindow(xwin->display_, xwin->wtwin_);
  XSelectInput(xwin->display_, xwin->wtwin_, KeyPressMask | KeyReleaseMask | ButtonPressMask);
  gc_values.graphics_exposures = False;
  xwin->gc_ = XCreateGC(xwin->display_, xwin->wtwin_, GCGraphicsExposures, &gc_values);  

  /* Create shared memory segment for direct framebuffer access */
  screen = DefaultScreen(xwin->display_);
  XGetWindowAttributes(xwin->display_, xwin->wtwin_, &win_attributes);
  localdepth=win_attributes.depth;


  // check if desired depth and X-Server bpp match
  if ( depth == 1 && localdepth != 8 ) { // BW-Images 
   fprintf (stderr, "%%setup_xwinbuffer(): depth=1 requires X-Server in 8 bpp mode!\n"); 
   return(-1); 
  } 			   
  else if ( depth == 2 && localdepth != 16 ) { // RGB565-Images 
   fprintf (stderr, "%%setup_xwinbuffer(): depth=2 requires X-Server in 16 bpp mode!\n"); 
   return(-1); 
  } 			   
  else if ( depth > 2 ) {
   fprintf (stderr, "%%setup_xwinbuffer(): not implemented depth/X-Server mode combination (%d, %d)!\n", depth, localdepth); 
   return(-1); 
  } 			   

  xwin->shminfo = (XShmSegmentInfo*) malloc (sizeof(XShmSegmentInfo));
  xwin->shmimage = XShmCreateImage(xwin->display_, DefaultVisual(xwin->display_, screen),
			     localdepth, ZPixmap, NULL, xwin->shminfo,
			     xwin->width,xwin->height);

  if ((xwin->shminfo->shmid=shmget(getpid(),
			  xwin->shmimage->bytes_per_line * xwin->shmimage->height,
			  IPC_CREAT | IPC_EXCL | 0777))==-1) {
   perror("shmget()");
   return(-1);
  } 			  
  xwin->shminfo->shmaddr = (char *)shmat(xwin->shminfo->shmid, (void *) 0, 0);
  xwin->shmimage->data = xwin->shminfo->shmaddr;
  XShmAttach(xwin->display_, xwin->shminfo);

  // Set color mapping
  if ( depth == 1 ) // BW-Images 
    desired_visual_template.class  = PseudoColor;
  else
    desired_visual_template.class = TrueColor; 

  desired_visual_template.screen = screen;
  visual_info_list = XGetVisualInfo(xwin->display_,  
				    VisualScreenMask | VisualClassMask, 
				    &desired_visual_template, 
				    &no_visual_matched); 
  if (no_visual_matched == 0) {
   fprintf(stderr, "No visual matched\n");
   return(-1);
  }


  // make color map 
  if ( depth == 1 ) { // color map needed when using BW-Images 
    wtcolormap = XCreateColormap(xwin->display_, 
                                 DefaultRootWindow(xwin->display_),
                                 visual_info_list[0].visual, AllocAll);
    numCols = 1<<localdepth;
    palette = (XColor *)malloc(numCols * sizeof(XColor));
    for (pixelNum=0; pixelNum < numCols; pixelNum++) {
        palette[pixelNum].pixel  = pixelNum;
        palette[pixelNum].red    = pixelNum << (16-localdepth);
        palette[pixelNum].green  = pixelNum << (16-localdepth);
        palette[pixelNum].blue   = pixelNum << (16-localdepth);
        palette[pixelNum].flags  = DoRed|DoGreen|DoBlue;
      }

    XStoreName(xwin->display_, xwin->wtwin_, prgname);
    XStoreColors(xwin->display_, wtcolormap, palette, numCols);
    XSetWindowColormap(xwin->display_, xwin->wtwin_, wtcolormap);
  }
  
  return(0);  
}


int sync_xwinbuffer(struct xwinbuffer *xwin)
{
 /* Draw screen onto display */
 XShmPutImage(xwin->display_, xwin->wtwin_, xwin->gc_, xwin->shmimage,
 	        0, 0, 0, 0, xwin->width, xwin->height, False);
 XSync(xwin->display_, 0); 
 return(0);
}

int keypoll_xwinbuffer (struct xwinbuffer *xwin, KeySym *key, int *button, int *mousex, int *mousey)
{
  XEvent event; 
 
  /* Check keyboard/mouse input */
  if (XPending(xwin->display_)>0) {
   *button=0;
   *key=0;
   XNextEvent(xwin->display_, &event);
   switch (event.type) {
    case ButtonPress:
     *mousex=event.xbutton.x;
     *mousey=event.xbutton.y;
     *button=1;
     break;
    case KeyPress:
     *key=XLookupKeysym((void *)&event,0);
     break;
    case MappingNotify:
     XRefreshKeyboardMapping((void  *)&event);
     break;
   }
   return(1);
  } else { 
   return(0);
  } 
}  

int close_xwinbuffer (struct xwinbuffer *xwin)
{
 /* Reset video-mode if it was changed */
 if (xwin->modeswitch) {
  videoModeReset(xwin);
 }

 /* Mark Shm as destroyed */
 shmctl(xwin->shminfo->shmid,IPC_RMID,NULL);

 /* Detatch shared memory segment */
 shmdt ((char *)xwin->shminfo->shmaddr);
 
 return(0);
}

/* Drawing utilities */

/* Clip to screen coordinates */
void ClipX(struct xwinbuffer *xwin, int *x)
{
 /* Clip X coordinates */
 if (*x<0) {
  *x=0;
 } else if (*x>xwin->maxx) {
  *x=xwin->maxx;
 } 
}

/* Clip to screen coordinates */
void ClipY(struct xwinbuffer *xwin, int *y)
{
 /* Clip Y coordinates */
 if (*y<0) {
  *y=0;
 } else if (*y>xwin->maxy) {
  *y=xwin->maxy;
 } 
}

void SortXY(int *a, int *b)
{
 int c;
 
 /* Swap if a>b */
 if (*a>*b) {
  c=*a;
  *a=*b;
  *b=c;
 }
}

/* Draw horizontal line into color buffer */
void hline (struct xwinbuffer *xwin, int x1, int x2, int y, int color)
{
 int i;
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;
 
 if (xwin==NULL) return; 
 /* Clip coordinates */
 ClipX(xwin,&x1);
 ClipX(xwin,&x2);
 SortXY(&x1,&x2);
 ClipY(xwin,&y);
 /* Draw */
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x1+y*xwin->width);
  for (i=0; i<(x2-x1+1); i++) {
   *curpos8=color8;
   curpos8++;
  }
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x1+y*xwin->width);
  for (i=0; i<=(x2-x1); i++) {
   *curpos16=color16;
   curpos16++;
  }
 }
}

/* Draw vertical line into color buffer */
void vline (struct xwinbuffer *xwin, int x, int y1, int y2, int color)
{
 int i;
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;

 if (xwin==NULL) return; 
 /* Clip coordinates */
 ClipX(xwin,&x);
 ClipY(xwin,&y1);
 ClipY(xwin,&y2);
 SortXY(&y1,&y2);
 /* Draw */
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   *curpos8=color8;
   curpos8 += xwin->width;
  }
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   *curpos16=color16;
   curpos16 += xwin->width;
  }
 }
}

/* Draw filled box */
void drawbox (struct xwinbuffer *xwin, int x1, int y1, int x2, int y2, int color)
{
 int i,j,step;
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;

 if (xwin==NULL) return; 
 /* Clip coordinates */
 ClipX(xwin,&x1);
 ClipX(xwin,&x2);
 SortXY(&x1,&x2);
 ClipY(xwin,&y1);
 ClipY(xwin,&y2);
 SortXY(&y1,&y2);

 /* Draw */
 step = xwin->width-(x2-x1);
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x1+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   for (j=x1; j<x2; j++) {
    *curpos8=color8;
    curpos8++;
   }
   curpos8 += step;
  }
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x1+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   for (j=x1; j<x2; j++) {
    *curpos16=color16;
    curpos16++;
   }
   curpos16 += step;
  }
 }
}

/* Draw character into buffer */
void drawchar (struct xwinbuffer *xwin, int x, int y, char c, int color)
{
 int i,j,k;
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;
 unsigned char *charpos;
 unsigned char bits[8]={128,64,32,16,8,4,2,1};
 unsigned char *bitpos;
 int ystep,xmax,ymax;

 if (xwin==NULL) return; 
 /* Clip coordinates */
 xmax=x+8;
 ymax=y+8;
 ClipX(xwin,&x);
 ClipY(xwin,&y);
 ClipX(xwin,&xmax);
 ClipY(xwin,&ymax);
 /* Variable setup */
 k = (unsigned char)c;
 k *= 8;
 charpos=font8x8;
 charpos += k;
 bitpos=bits;
 ystep=xwin->width-(xmax-x);
 /* Draw */
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x+y*xwin->width);
  for (i=y; i<ymax; i++) {
   bitpos=bits;
   for (j=x; j<xmax; j++) {
    if ((*charpos & *bitpos)==*bitpos) {
     *curpos8=color8;
    } 
    curpos8++;
    bitpos++;
   }
   curpos8 += ystep;
   charpos++;
  }
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x+y*xwin->width);
  for (i=y; i<ymax; i++) {
   bitpos=bits;
   for (j=x; j<xmax; j++) {
    if ((*charpos & *bitpos)==*bitpos) {
     *curpos16=color16;
    } 
    curpos16++;
    bitpos++;
   }
   curpos16 += ystep;
   charpos++;
  }
 }
}

/* Draw string into buffer */
void drawstring (struct xwinbuffer *xwin, int x, int y, char *s, int color)
{
 int i,length;
 char *curchar;
 int curx;

 length=strlen(s);
 curchar=s;
 curx=x;
 for (i=0; i<length; i++) {
  drawchar(xwin,curx,y,*curchar,color);
  curchar++;
  curx += 8;
 } 
}

void drawpixel(struct xwinbuffer *xwin, int x, int y, int color)
{
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;

 /* Clip coordinates */
 ClipX(xwin,&x);
 ClipY(xwin,&y);
 /* Draw */
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x+y*xwin->width);
  *curpos8=color8;
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x+y*xwin->width);
  *curpos16=color16;
 }
}

/* Draw pixel into 8bpp screen without clipping */
void drawpixel8fast(struct xwinbuffer *xwin, int x, int y, int color)
{
 unsigned char *curpos8;

 curpos8 = (unsigned char *)xwin->shmimage->data;
 curpos8 += (x+y*xwin->width);
 *curpos8=(unsigned char)color;
}

/* Draw pixel into 16bpp screen without clipping */
void drawpixel16fast(struct xwinbuffer *xwin, int x, int y, int color)
{
 unsigned short *curpos16;
 
 curpos16 = (unsigned short *)xwin->shmimage->data;
 curpos16 += (x+y*xwin->width);
 *curpos16=(unsigned short)color;
}

#define ABS(a) (((a)<0) ? -(a) : (a))

/* Draw a line between two points */
void drawline(struct xwinbuffer *xwin, int x1, int y1, int x2, int y2, int color)
{
 int dx;
 int dy;
 int ax;
 int ay;
 int sx;
 int sy;

 int x = x1;
 int y = y1;

 ClipX(xwin,&x1);
 ClipY(xwin,&y1);
 ClipX(xwin,&x2);
 ClipY(xwin,&y2);

 dx = x2 - x1;
 dy = y2 - y1;
 ax = ABS(dx) << 1;
 ay = ABS(dy) << 1;
 sx = (dx >= 0) ? 1 : -1;
 sy = (dy >= 0) ? 1 : -1;
 x = x1;
 y = y1;

 if (xwin->depth==1) {
  if (ax > ay) {
	int d = ay - (ax >> 1);
	while (x != x2) {

	    drawpixel8fast(xwin, x, y, color);

	    if (d > 0 || (d == 0 && sx == 1)) {
		y += sy;
		d -= ax;
	    }
	    x += sx;
	    d += ay;
	}
    } else {
	int d = ax - (ay >> 1);
	while (y != y2) {
	    drawpixel8fast(xwin, x, y, color);

	    if (d > 0 || (d == 0 && sy == 1)) {
		x += sx;
		d -= ay;
	    }
	    y += sy;
	    d += ax;
	}
    }
    drawpixel8fast(xwin, x, y, color);
 } else if (xwin->depth==2) {
  if (ax > ay) {
	int d = ay - (ax >> 1);
	while (x != x2) {

	    drawpixel16fast (xwin, x, y, color);

	    if (d > 0 || (d == 0 && sx == 1)) {
		y += sy;
		d -= ax;
	    }
	    x += sx;
	    d += ay;
	}
    } else {
	int d = ax - (ay >> 1);
	while (y != y2) {
	    drawpixel16fast (xwin, x, y, color);

	    if (d > 0 || (d == 0 && sy == 1)) {
		x += sx;
		d -= ay;
	    }
	    y += sy;
	    d += ax;
	}
    }
    drawpixel16fast (xwin, x, y, color);
 } // depth check 
}


/* Change the current X11 video mode */
void videoModeChange(struct xwinbuffer *xwin, int width, int height)
{
 int i;

 XF86VidModeGetAllModeLines(xwin->display_,XDefaultScreen(xwin->display_), &xwin->vm_count,&xwin->vm_modelines);
 for (i = 0; i < xwin->vm_count; i++) {
  if (xwin->vm_modelines[i]->hdisplay == width && xwin->vm_modelines[i]->vdisplay == height) {
   xwin->vm_smode=i;
  }
 }

 if (xwin->vm_smode != -1) {
  XF86VidModeSwitchToMode(xwin->display_,XDefaultScreen(xwin->display_),xwin->vm_modelines[xwin->vm_smode]);
  XFlush(xwin->display_);
  xwin->modeswitch=1;
 }
}

/* Reset X11 video mode */
void videoModeReset(struct xwinbuffer *xwin)
{
 if (xwin->vm_smode != -1) {
  XF86VidModeSwitchToMode(xwin->display_,XDefaultScreen(xwin->display_),xwin->vm_modelines[0]);
  XFlush(xwin->display_);
 }
 xwin->vm_smode=-1;
 xwin->modeswitch=0;
}
